KeyValueStore

Trait KeyValueStore 

pub trait KeyValueStore: Send + Sync {
    // Required methods
    fn get_all(&self) -> HashMap<String, Value>;
    fn get(&self, key: &str) -> Option<Value>;
    fn set(&self, key: &str, value: Value);
    fn mset(&self, kv_pairs: &[(String, Value)]);
    fn exists(&self, key: &str) -> bool;
    fn keys(&self) -> Vec<String>;
    fn mget(&self, keys: &[String]) -> HashMap<String, Value>;
    fn delete(&self, key: &str);
    fn get_prefix(&self, prefix: &str) -> HashMap<String, Value>;
}
Abstract interface for key-value storage backends

Required Methods

fn get_all(&self) -> HashMap<String, Value>

Get all key-value pairs

fn get(&self, key: &str) -> Option<Value>

Get value for key

fn set(&self, key: &str, value: Value)

Set value for key

fn mset(&self, kv_pairs: &[(String, Value)])

fn exists(&self, key: &str) -> bool

Check if key exists

fn keys(&self) -> Vec<String>

Get all keys matching pattern

fn mget(&self, keys: &[String]) -> HashMap<String, Value>

Batch get multiple keys, returns map with found key-value pairs

fn delete(&self, key: &str)

Delete a key

fn get_prefix(&self, prefix: &str) -> HashMap<String, Value>

Get all key-value pairs with keys starting with the given prefix

Implementors